<?php
/**
 * @file
 * Drupal Inspect module base class
 */

/**
 * Inspect base class
 *
 * @uses PHP:ini:precision
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class Inspect {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Maximum sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_MAX = 10;

  /**
   * Default sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 10;

  /**
   * Minimum string truncation.
   *
   * @type integer
   */
  const TRUNCATE_MIN = 10000;

  /**
   * Default string truncation.
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 1000;

  /**
   * Will be substracted from current logging output max length, to allow for insert query and inaccuracies of length measuring when inspecting.
   *
   * @type integer
   */
  const LOG_MAX_MARGIN = 10000;

  /**
   * @type integer
   */
  const ERROR_ALGORITHM = 100;

  /**
   * @type integer
   */
  const ERROR_USER = 101;

  /**
   * @type integer
   */
  const ERROR_OUTPUTLENGTH = 102;

  /**
   * @type integer
   */
  const ERROR_EXECTIME = 103;

  /**
   * Default inspection options.
   *
   *   Various options:
   *   - (array) category: logging category
   *   - (boolean) file_line: display file and line of call to inspect function/method
   *   - (array|NULL) filter: filter out that|those key name(s)
   *   - (boolean) one_lined: (get only) for creating one-lined formatting instead of multi-lined
   *
   *   String options:
   *   - (array) needles: list of char|string needles to be used in str_replace
   *   - (array) replacers: list of char|string replacers to be used in str_replace
   *   - (string) quote_begin: char|string used as initial quote
   *   - (string) quote_end: char|string used as final quote
   *
   *   Formatting options:
   *   - (array) delimiters: between object/array buckets, and for InspectTrace between frames and frame buckets
   *   - (string) pre_indent: indentation of object/array buckets
   *   - (string) indent: indentation of object/array buckets (gets multiplied by current depth)
   *   - (string) enclose_tag: html enclosure tag
   *   - (string) newline: newline character
   *
   * @see Inspect::_options()
   * @var array $_options
   */
  protected static $_options = array(
    'get' => array(
      'category' => 'inspect',
      'severity' => WATCHDOG_DEBUG,
      'file_line' => TRUE,
      'filter' => NULL,
      'one_lined' => FALSE,
      'needles' => array( "\n", "\r", "\t", '<', '>', '"', "'" ),
      'replacers' => array( '_NL_', '_CR_', '_TB_', '&#60;', '&#62;', '&#34;', '&#39;' ),
      'quote_begin' => '&#62;&#62;',
      'quote_end' => '&#60;&#60;',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n"  // Before last bucket.
      ),
      'pre_indent' => '',
      'indent' => '.  ',
      'enclose_tag' => 'pre',
      'newline' => "\n",
    ),
    'log' => array(
      'category' => 'inspect',
      'severity' => WATCHDOG_DEBUG,
      'file_line' => TRUE,
      'filter' => NULL,
      'one_lined' => FALSE,
      'needles' => array( "\n", "\r", "\t", '<', '>', '"', "'" ),
      'replacers' => array( '_NL_', '_CR_', '_TB_', '&#60;', '&#62;', '&#34;', '&#39;' ),
      'quote_begin' => '&#62;&#62;',
      'quote_end' => '&#60;&#60;',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n"  // Before last bucket.
      ),
      'pre_indent' => '',
      'indent' => '.  ',
      'enclose_tag' => 'pre',
      'newline' => "\n",
    ),
    //  syslog: Uses + for delimiters and newlines.
    'syslog' => array(
      'category' => 'inspect',
      'severity' => WATCHDOG_DEBUG,
      'file_line' => TRUE,
      'filter' => NULL,
      'one_lined' => FALSE,
      'needles' => array( "\n", "\r", "\t", '<', '>', '"', "'" ),
      'replacers' => array( '_NL_', '_CR_', '_TB_', '_LT_', '_GT_', '_QT_', '_AP_' ),
      'quote_begin' => '"',
      'quote_end' => '"',
      'delimiters' => array(
        '+', // Before first bucket.
        '+', // Before later bucket.
        '+'  // Before last bucket.
      ),
      'pre_indent' => '',
      'indent' => '. ',
      'enclose_tag' => '',
      'newline' => '+',
    ),
    'file' => array(
      'category' => 'inspect',
      'severity' => WATCHDOG_DEBUG,
      'file_line' => TRUE,
      'filter' => NULL,
      'one_lined' => FALSE,
      'needles' => array( "\r", "\t" ),
      'replacers' => array( "_CR_\r", "_TB_\t" ),
      'quote_begin' => '>>',
      'quote_end' => '<<',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n"  // Before last bucket.
      ),
      'pre_indent' => '',
      'indent' => '.  ',
      'enclose_tag' => '',
      'newline' => "\n",
    ),
  );

  /**
   * Options overriding the default, when truthy option one_lined.
   *
   * Apart form formatting, also makes default depth 1 and call file and line default off.
   *
   * @var array $_options_oneLined
   */
  protected static $_options_oneLined = array(
    'file_line' => FALSE,
    'delimiters' => array(
      '', // Before first bucket.
      ', ', // Before later bucket.
      ''  // Before last bucket.
    ),
    'indent' => '',
    'enclose_tag' => '',
    'newline' => ' ',
    'no_preface' => TRUE,
    'name' => 'none',
  );

  /**
   * Options that may be set by value as well as by key.
   *
   * @var array $_options_byValue
   */
  protected static $_options_byValue = array('hide_scalars', 'hide_paths', 'by_user', 'one_lined', 'no_folding', 'no_preface');

  /**
   * Inspect session number, page load number, request number.
   *
   * @see Inspect::_initSessionCounters()
   * @see Inspect::_updatePageLoadNumber()
   * @var array $_sessionCounters
   */
  protected static $_sessionCounters = array(
      'n/a',
      0,
      1
  );

  /**
   * @see Inspect::_permit()
   * @var array $_permissions
   */
  private static $_permissions = array(
    'inspect log' => NULL, // NULL ~ not checked yet | boolean ~ checked.
    'inspect get' => NULL,
  );

  /**
   * @see Inspect::_init()
   * @var boolean|NULL $_init
   */
  protected static $_init;

  /**
   * Length of document root or real path (the shorter one of them).
   *
   * @see Inspect::_init()
   * @var integer $_pathLength
   */
  protected static $_pathLength;

  /**
   * Document root and - if symlinked root real path (__FILE__ vs. SCRIPT_FILENAME) - with trailing slash, and as regular expression(s) (/^.../)
   *
   * @see Inspect::_init()
   * @var array $_paths
   */
  protected static $_paths;

  /**
   * Request time start in milliseconds.
   *
   * @see Inspect::requestTimeMilli()
   * @see Inspect::_init()
   * @var float $_requestTimeMilli
   */
  protected static $_requestTimeMilli = 0;

  /**
   * Maximum output length when logging.
   *
   * @see Inspect::_init()
   * @var integer $_outputMax_log
   */
  protected static $_outputMax_log = 0;

  /**
   * Maximum output length when getting.
   *
   * @see Inspect::_init()
   * @var integer $_outputMax_get
   */
  protected static $_outputMax_get = 0;

  /**
   * Maximum output length for current mode (log or get/file).
   *
   * @var integer $_outputMax
   */
  protected static $_outputMax = 0;

  /**
   * String length of an inspection output a any time during inspection.
   *
   * NB: Must be reset to zero every time a toplevel method initiates an operation.
   *
   * @var integer $_outputLength
   */
  protected static $_outputLength = 0;

  /**
   * PHP max_execution_time as measured at first call to _inspect().
   *
   * @see Inspect::_init()
   * @var integer $_maxExecTime
   */
  protected static $_maxExecTime = 0;

  /**
   * _inspect() aborts when called this percent of max_execution_time has passed.
   *
   * @see Inspect::_init()
   * @var integer $_maxExecTimePercent
   */
  protected static $_maxExecTimePercent = 0;

  /**
   * _inspect() will abort if called later than this time ($_maxExecTimePercent percent of max execution time).
   *
   * @see Inspect::_init()
   * @var integer $_maxExecTimeout
   */
  protected static $_maxExecTimeout = 0;

  /**
   * System logs to syslog?
   *
   * @see Inspect::_init()
   * @var boolean|NULL $_syslog
   */
  protected static $_syslog;

  /**
   * Times _inspect() has been called.
   *
   * @see Inspect::_inspect()
   * @var integer $_callNo
   */
  protected static $_callNo = 0;

  /**
   * Log number in current request.
   *
   * NB: Any logging or filing toplevel method must increase this var.
   *
   * @var integer $_logNo
   */
  protected static $_logNo = 0;

  /**
   * private://module/inspect
   *
   * @see Inspect::_ensureDir()
   * @var string|NULL|boolean $_dir
   */
  protected static $_dir;

  /**
   * @see Inspect::_ensureDir()
   * @var string|NULL $_logDir
   */
  protected static $_logDir;

  /**
   * Establish session number, page load number and request number.
   *
   * For hook_init() implementation.
   *
   * @param array $counters
   * @return void
   */
  public static function _initSessionCounters() {
    static $called;
    if (!$called) {
      $called = TRUE;
      if (isset($_COOKIE['inspect__sc']) && preg_match('/^[a-zA-Z\d]+\:\d{1,5}\:\d{1,5}$/', $_COOKIE['inspect__sc'])) {
        $c = explode(':', $_COOKIE['inspect__sc']);
        self::$_sessionCounters = array(
          $c[0], // Session number.
          $c[1], // Page load number.
          (int)$c[2] + 1 // Request number.
        );
      }
      else {
        $c = explode('.', uniqid('', TRUE));
        self::$_sessionCounters = array(
          Inspect::baseConvert($c[0], 16, 62) . Inspect::baseConvert($c[1], 16, 62),
          0, // Will be increased when page delivery is guaranteed (hook_preprocess_html).
          1
        );
      }
      setcookie(
        'inspect__sc', join(':', self::$_sessionCounters), 0,
        $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], FALSE // httponly
      );
    }
  }

  /**
   * Increases page load number and sets updated session counter cookie.
   *
   * For hook_preprocess_html() implementation.
   *
   * @return void
   */
  public static function _updatePageLoadNumber() {
    static $called;
    if (!$called) {
      $called = TRUE;
      ++self::$_sessionCounters[1];
      setcookie(
        'inspect__sc', join(':', self::$_sessionCounters), 0,
        $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], FALSE // httponly
      );
    }
  }

  /**
   * Checks if user is allowed to output to that target.
   *
   *  Special rules:
   *  - filing uses same permissions as logging
   *  - drush/CLI is always allowed to get|log|file
   *  - cron may be allowed to log (admin option overrides)
   *
   * @param string $target
   *  - default: 'get'
   *  - values: 'get' | 'log' | 'file'
   * @return boolean
   */
  protected static function _permit($target = 'get') {
    switch ('' . $target) {
      case 'log':
      case 'file':
        if (($perm = self::$_permissions['inspect log']) || $perm === FALSE) {
          return $perm;
        }
        //  ~ NULL: Was not initialized.
        $perm = FALSE;
        if (user_access('inspect log')
          || drupal_is_cli() // drush/CLI always allowed to log.
          || ( // Cron may be allowed to log.
            !$GLOBALS['user']->uid // anonymous user
            && variable_get('inspect_allow_log_cron', FALSE)
            && !empty($GLOBALS['_SERVER']) && !empty($_SERVER['SCRIPT_NAME']) && $_SERVER['SCRIPT_NAME'] === '/cron.php'
            && ($cron_key = variable_get('cron_key', ''))
            && !empty($GLOBALS['_GET']) && !empty($_GET['cron_key']) && $_GET['cron_key'] === $cron_key
          )
        ) {
          $perm = TRUE;
        }
        return self::$_permissions['inspect log'] = $perm;
      default: // get
        if (($perm = self::$_permissions['inspect get']) || $perm === FALSE) {
          return $perm;
        }
        //  ~ NULL: Was not initialized.
        $perm = FALSE;
        //  drush/CLI always allowed to get.
        if (drupal_is_cli() || user_access('inspect get')) {
          $perm = TRUE;
        }
        return self::$_permissions['inspect get'] = $perm;
    }
    return FALSE;
  }

  /**
   * @return boolean
   */
  protected static function _init() {
    //  Find real path, and document root (if possible), for removal from absolute paths.
    $le0 = drupal_strlen($realPath = DRUPAL_ROOT); // getcwd().
    $le1 = !empty($GLOBALS['_SERVER']) && !empty($_SERVER['SCRIPT_FILENAME']) ?
        drupal_strlen($docRoot = dirname($_SERVER['SCRIPT_FILENAME'])) : 0;
    //  Find longest path.
    $osNix = DIRECTORY_SEPARATOR == '/';
    self::$_pathLength = ($le0 < $le1 ? $le0 : $le1) - ($osNix ? 0 : 2); // And remove drive C:, for Windows.
    self::$_paths = array(
      $osNix ? ('/^' . preg_quote($realPath, '/') . '\//') :
        ('/^(' . $realPath{0} . '\:)?' . str_replace('\\\\', '[\\\\\/]', preg_quote(drupal_substr($realPath, 2))) . '[\\\\\/]/i')
    );
    if ($le1 && $docRoot != $realPath) {
      self::$_paths[] = $osNix ? ('/^' . preg_quote($docRoot, '/') . '\//') :
        ('/^(' . $docRoot{0} . '\:)?' . preg_quote(drupal_substr($docRoot, 2), '/') . '\//i');
    }

    //  Establish output max lengths.
    self::outputMax();

    //  Establish request time start.
    $t = self::requestTimeMilli();

    //  Establish max execution time abort.
    self::$_maxExecTimeout = floor($t / 1000)
      + floor(
        (self::$_maxExecTime = ini_get('max_execution_time'))
        * ((self::$_maxExecTimePercent = variable_get('inspect_exectime_percent', 70)) / 100)
      );

    //  Prepare default options - make log options equal syslog options if syslog enabled, and get options equal file options if drush/CLI request.
    $o =& self::$_options;
    //  If syslog: use syslog options as log options.
    if ((self::$_syslog = module_exists('syslog'))) {
      $o['log'] = $o['syslog'];
    }
    //  If drush/CLI: use default file options instead of get options.
    if (drupal_is_cli()) {
      $o['get'] = $o['file'];
    }

    return (self::$_init = TRUE);
  }

  /**
   * @see Inspect::$_options
   * @param string $target
   * @param array &$defaultOptions
   * @param mixed $options
   * @param string $integer
   *  - default: 'depth'
   *  - what to interpretate $options as, if $options is an integer
   * @param string $boolean
   *  - default: '' (~ ignore $options if boolean)
   *  - what to interpretate $options as, if $options is a boolean
   * @return array
   */
  protected static function _options($target, &$defaultOptions, $options, $integer = 'depth', $boolean = '') {
    $hidePaths = FALSE;
    switch ('' . $target) {
      case 'log':
      case 'file':
        break;
      default:
        $target = 'get'; // Safe-guard against bad $target arg.
        $hidePaths = TRUE; // We always hide paths when getting.
    }
    //  Resolve non-array overriding options.
    $isArr = $checkOpts = FALSE;
    $opts = NULL;
    switch (gettype($options)) {
      case 'NULL':
        break;
      case 'array':
        $isArr = $checkOpts = TRUE;
        $opts = $options;
        break;
      case 'object':
        $checkOpts = TRUE;
        $opts = get_object_vars($options);
        break;
      case 'string':
        if ($options !== '') {
          $opts = array('message' => $options);
        }
        break;
      case 'integer':
        $opts = array($integer => $options);
        break;
      case 'boolean':
        if ($boolean) {
          $opts = array($boolean => $options);
        }
        else { // Ignore.
          $opts = NULL;
        }
        break;
      default:
        //  Ignore argument.
        $opts = NULL;
    }

    //  Load some default values.
    $depthOpt = static::DEPTH_DEFAULT;
    $truncOpt = static::TRUNCATE_DEFAULT;
    $hideScalars = FALSE;
    $oneLined = FALSE;

    if ($opts) {
      $needles = $replacers = NULL;
      //  We do note care to check options like this if only one and is message or the integer (depth|trace limit) option.
      if ($checkOpts) {
        //  If array, some options may also be set as value (and thus numeric key) ~ interpreted as truthy boolean.
        $byValue = self::$_options_byValue;
        foreach ($byValue as $v) {
          if (($index = array_search($v, $opts, TRUE)) !== FALSE) {
            $opts[$v] = TRUE;
            unset($opts[$index]);
          }
        }

        //  Check validity of overriding options.
        if (isset($opts['needles'])) {
          if (($v = $opts['needles']) && is_array($v)) {
            $needles = $v;
          }
          unset($opts['needles']);
        }
        if (isset($opts['replacers'])) {
          if (($v = $opts['replacers']) && is_array($v)) {
            $replacers = $v;
          }
          unset($opts['replacers']);
        }
        if ($target == 'get' && !empty($opts['one_lined'])) {
          $oneLined = TRUE;
          $depthOpt = 1;
        }
        else {
          unset($opts['one_lined']);
        }
      }
      //  Read depth, hide scalars, truncate, filter, hide paths, if they exist.
      if (isset($opts['depth'])) {
        $depthOpt = (int)$opts['depth'];
      }
      if (!$hideScalars && !empty($opts['hide_scalars'])) {
        $hideScalars = TRUE;
      }
      if (isset($opts['truncate'])) {
        if (($v = (int)$opts['truncate'])) {
          if ($v < 0) {
            $v = 0;
          }
          elseif ($v > static::TRUNCATE_MIN) { // If suggested truncate is more than overall min.
            $v = static::TRUNCATE_MIN;
          }
        }
        $truncOpt = $v;
      }
      if (isset($opts['filter'])) {
        if (is_array($v = $opts['filter'])) {
          if (!$v) {
            unset($opts['filter']);
          }
        }
        elseif (('' . $v) !== '') {
          $opts['filter'] = array(0 => '' . $v);
        }
      }
      //  We set the bucket hide_paths no matter what, and it can only override in negative (safer) manner.
      if (!$hidePaths && !empty($opts['hide_paths'])) {
        $hidePaths = TRUE;
      }

      //  Load defaults, and make arg options override.
      $opts = !$oneLined ? array_merge($defaultOptions, $opts) : array_merge($defaultOptions, static::$_options_oneLined, $opts);
      if ($needles) {
        $opts['needles'] = array_merge($opts['needles'], $needles);
      }
      if ($replacers) {
        $opts['replacers'] = array_merge($opts['replacers'], $replacers);
      }
    }
    else {
      $opts = $defaultOptions; // Copy, it's an array.
    }

    //  Round it up.
    $opts['target'] = $target;
    //  Check depth.
    if ($depthOpt < 0) {
      $depthOpt = 0;
    }
    elseif ($depthOpt && $depthOpt > static::DEPTH_MAX) {
      $depthOpt = static::DEPTH_MAX;
    }
    //  Check severity.
    if ($opts['severity'] < 0) {
      $opts['severity'] = WATCHDOG_DEBUG;
    }
    //  Load resolved values of depth and truncate into the options array.
    $opts['depth'] = $depthOpt;
    $opts['hide_scalars'] = $hideScalars;
    $opts['truncate'] = $truncOpt;
    $opts['hide_paths'] = $hidePaths;
    //  Escape name and message options
    if (!empty($opts['name'])) {
      //  Dont escape if target is file.
      $u = $target != 'file' ? htmlspecialchars(
        $opts['name'],
        ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
        'UTF-8',
        FALSE // No double encoding
      ) : $opts['name'];
      $opts['name'] = strlen($u) <= 255 ? $u : substr($u, 0, 255); // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
    }
    else {
      $opts['name'] = '';
    }
    if (!empty($opts['message'])) {
      //  Dont escape if target is file.
      $u = $target != 'file' ? htmlspecialchars(
        $opts['message'],
        ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
        'UTF-8',
        FALSE // No double encoding
      ) : $opts['name'];
      $opts['message'] = strlen($u) <= 255 ? $u : substr($u, 0, 255); // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
    }
    return $opts;
  }

  /**
   * The actual variable analyser.
   *
   *  Caller must - prior to calling this method - guarantee to call:
   *   - _permit()
   *   - _init(), unless self::$_init is true
   *   - _options()
   *
   *  Array/object value of $options (any number of options):
   *   - (integer) depth (default 10, max 10 ~ Inspect::DEPTH_DEFAULT|DEPTH_MAX)
   *   - (integer) truncate (default 1000, max 10000 ~ Inspect::TRUNCATE_MIN|TRUNCATE_DEFAULT)
   *   - (bool) hide_scalars: hide values of scalar vars, and type of resource (also supported as value 'hide_scalars')
   *   - (bool) hide_paths: only relevant for log (paths always hidden for get) (also supported as value 'hide_paths')
   *   - (string) name: '$var_name' or "\$var_name", must be escaped, will be truncated to 255 (default empty, use 'none' to omit)
   *   - (string) message, will be truncated to 255 (default empty)
   *   - (string) category: logging category (default inspect)
   *   - (integer) severity: default WATCHDOG_DEBUG
   *   - (string|array) filter: filter out that|those key name(s) (default empty)
   *   - (array) needles: additional list of char|string needles to be used in str_replace
   *   - (array) replacers: additional list of char|string replacers to be used in str_replace
   *   - (boolean) one_lined: create one-lined formatting (getting only, and also supported as value 'one_lined')
   *   - (boolean) no_folding: dont make output expansible/collapsible (also supported as value 'no_folding')
   *   - (boolean) file_line: display file and line of call to inspect function/method (default do)
   *   - (bool) no_preface: truthy to omit [Inspect... preface
   *   - (string) quote_begin: char|string used as initial quote (default >>)
   *   - (string) quote_end: char|string used as final quote (default <<)
   *   - (array) delimiters: between object/array buckets (default newlines)
   *   - (string) indent: indentation of object/array buckets (default 2 spaces)
   *   - (string) enclose_tag: html enclosure tag for the log as a whole (default pre)
   *   - (string) newline: newline character (default newline)
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes paths at the beginning of variables:
   *   - always removes document root
   *   - always removes relative and absolute paths when getting (optional when logging)
   *   - path removal takes place before escaping and truncating (except for overly long strings)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - always stops at 10th (Inspect::DEPTH_MAX) recursion or less (~ the depth option)
   *   - when buckets of an object/array is analyzed, buckets of buckets will be analyzed to one lesser depth
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * @see Inspect::_permit()
   * @see Inspect::_init()
   * @see Inspect::_options()
   * @throws Exception
   *  - UNCAUGH, if cumulate inspection output length exceeds maximum
   * @param mixed $var
   * @param array &$options
   *   - by reference, but will not be altered
   * @param integer $_curDepth
   *   - default: zero
   *   - recursion counter, do not call it with anything but zero
   * @return string
   */
  protected static function _inspect($var, &$options, $_curDepth = 0) {
    //  Execution time check, every 1000th time.
    if (!((++self::$_callNo) % 1000) && time() > self::$_maxExecTimeout) {
      throw new Exception(
        'Inspection aborted: was called after ' . self::$_maxExecTimePercent . '% of PHP max_execution_time['
          . self::$_maxExecTime . '] had passed, using options depth[' . $options['depth']
          . '] and truncate[' . $options['truncate'] . '], try less depth or more truncation.',
        Inspect::ERROR_EXECTIME
      );
    }
    //  Output length check.
    if (($u = self::$_outputLength) && $u > self::$_outputMax) {
      throw new Exception(
        'Inspection aborted: output length ' . $u . ' exceeds maximum ' . self::$_outputMax . ', using options depth[' . $options['depth']
          . '] and truncate[' . $options['truncate'] . '], try less depth or more truncation.',
        Inspect::ERROR_OUTPUTLENGTH
      );
    }
    $maxDepth = $options['depth'];
    //  Recursion limit check.
    if ($_curDepth && $_curDepth > $maxDepth) {
      throw new Exception(
        'Algo error, current depth ' . $_curDepth . ' exceeds max depth ' . $maxDepth . '.',
        Inspect::ERROR_ALGORITHM
      );
    }
    //  Object or array.
    $tArr = FALSE;
    if (is_object($var) || ($tArr = is_array($var))) {
      $s = $tArr ? 'array' : ($tObj = get_class($var));
      if (!$_curDepth && !strlen($options['name'])) {
        $options['name'] = $s;
      }
      $s = '(' . $s . ':';
      //  If at max depth: simply get length of the container.
      if ($_curDepth == $maxDepth) {
        if ($tArr) {
          $s .= ($nSubs = count($var)) . ')' . (!$nSubs ? '' : ' [...]');
        }
        else { // stdClass cannot be counted, simply iterate
          $nSubs = 0;
          foreach ($var as $sub) {
            ++$nSubs;
          }
          $s .= $nSubs . ')' . (!$nSubs ? '' : ' {...}');
        }
        self::$_outputLength += strlen($s); // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
        return $s;
      }
      //  Dive into container buckets.
      else {
        $delimFirst = $options['delimiters'][0]
          . $options['pre_indent'] . str_repeat($indent = $options['indent'], $_curDepth + 1);
        $delimLater = $options['delimiters'][1]
          . $options['pre_indent'] . str_repeat($indent, $_curDepth + 1);
        $nSubs = 0;
        $sSub = '';
        //  Different iterations for object vs array, because:
        //  - array ust iterate by explicit reference,
        //    whereas object may fail fatally doing that (An iterator cannot be used with foreach by reference)
        //  - they must have different self reference checks
        //  - array must be checked for (all) numeric indices
        if ($tArr) {
          $numArr = TRUE; // if all numeric indices, use [] instead of {}
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $s
          );
          foreach ($var as $key => &$sub) { // by explicit reference
            $delim = (++$nSubs) > 1 ? $delimLater : $delimFirst;
            //  GLOBALS check, skip self reference.
            if ($key === 'GLOBALS' && is_array($sub) && array_key_exists('GLOBALS', $sub)) {
              $numArr = FALSE;
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . 'GLOBALS: (array) *RECURSION*'
              );
              $sSub .= $u;
              continue;
            }
            if ($numArr && $key !== $nSubs - 1) { // numeric array check
              $numArr = FALSE;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . $key . ': F'
              );
              $sSub .= $u;
            }
            else {
              $options['key'] = $key; // pass key, for password check
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . $key . ': '
              );
              $sSub .= $u . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $u = $nSubs . ') '
              . ($numArr ? '[' : '{')
          );
          $s .= $u . $sSub;
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $u = $options['delimiters'][2] . str_repeat($indent, $_curDepth)
              . ($numArr ? ']' : '}')
          );
          return $s . $u;
        }
        else { // object
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
              $s
          );
          foreach ($var as $key => $sub) { // not &$sub, iterating object is implicitly by reference
            $delim = (++$nSubs) > 1 ? $delimLater : $delimFirst;
            if ($sub === $var) { // check for identity (self reference)
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . $key . ': (' . $tObj . ') *RECURSION*'
              );
              $sSub .= $u;
              continue;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . $key . ': F'
              );
              $sSub .= $u;
            }
            else {
              $options['key'] = $key; // pass key, for password check
              self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
                $u = $delim . $key . ': '
              );
              $sSub .= $u . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $u = $nSubs . ') {'
          );
          $s .= $u . $sSub;
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $u = $options['delimiters'][2] . str_repeat($indent, $_curDepth)
              . '}'
          );
          return $s . $u;
        }
      }
    }
    //  Scalars, NULL and resource.
    switch (($t = gettype($var))) {
      case 'boolean':
        self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $s = '(' . $t . ')' . (!$options['hide_scalars'] ? ($var ? ' TRUE' : ' FALSE') : '')
        );
        return $s;
      case 'integer':
      case 'double':
      case 'float':
        if (!$options['hide_scalars'] && !is_finite($var)) {
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $s = '(' . (is_nan($var) ? 'NaN' : 'infinite') . ')'
          );
          return $s;
        }
        self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $s = '(' . ($t == 'double' ? 'float' : $t) . ')'
            . ( $options['hide_scalars'] ? '' :
              (!$var ? ' 0' : (' ' . self::numberToString($var)) ) )
        );
        return $s;
      case 'string':
        //  (string:n0|n1|n2): n0 ~ multibyte length, n1 ~ ascii length, n2 only occurs if truncation.
        $s = '(string:';
        //  Get multibyte length and ascii length, and check for invalid UTF-8.
        $leRaw = strlen($var); // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
        if (!drupal_validate_utf8($var)) {
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $s .= '?|' . $leRaw . '|0) *INVALID_UTF8*'
          );
          return $s;
        }
        $leMb = $leRaw ? drupal_strlen($var) : 0;
        $s .= !$leRaw ? '0|0' : ($leMb . '|' . $leRaw);
        //  If empty or hide scalar values.
        if ($options['hide_scalars'] || !$leRaw) {
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $s .= ')'
              . ( $options['hide_scalars'] ? '' : // no quotes if hide value
                (' ' . $options['quote_begin'] . $options['quote_end']) )
          );
          return $s;
        }
        //  If very long string, truncate to absolute max before working on the string internals.
        $truncedTo = 0;
        if ($leMb > ($trunc = $options['truncate']) && $leMb > ($trunc_min = static::TRUNCATE_MIN)) {
          $var = drupal_substr($var, 0, $trunc_min);
          $truncedTo = $leMb = $trunc_min;
        }
        //  If truncate is zero.
        if (($pw = ($_curDepth > 0 && ($options['key'] === 'pass' || $options['key'] === 'password'))) || $trunc < 1) {
          //  Look for stringed integer (consists solely of digits), but only if not truncated.
          if (!$pw && !$truncedTo && preg_match('/^\d+$/', $var)) {
            self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
              $s .= ') ' . $options['quote_begin'] . $var . $options['quote_end'] // Stringed integer.
            );
            return $s;
          }
          self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
            $s .= '|0) ' . $options['quote_begin'] . '...' . $options['quote_end'] // |0 ~ flag truncation
          );
          return $s;
        }
        //  Remove document root (always) and/or path.
        //  Do it before escaping (needles may contain dot or slash).
        $pathRem = FALSE;
        if ($leRaw > 3) {
          if ($options['hide_paths']) { //  This algo also removes document root.
            if (DIRECTORY_SEPARATOR == '/') { // *nix; non-Windows.
              if ($var{0} === '/' // Absolute path.
                || ( $var{0} === '.' && (strpos($var, '../') === 0 || strpos($var, './') === 0) ) // Relative.
              ) {
                $pathRem = TRUE;
              }
            }
            else { // Windows.
              if ( strpos($var, '/') !== FALSE || strpos($var, '\\') ) { // Has slash.
                if (($char = $var{0}) === '.') { // Check for relative.
                  if (strpos($var, '../') === 0 || strpos($var, './') === 0
                    || strpos($var, '..\\') === 0 || strpos($var, '.\\') === 0) {
                    $pathRem = TRUE;
                  }
                }
                //  Absolute path (ascii letter + : + slash).
                //elseif (preg_match('/^[A-Za-z]\:[\/\x5C]/', $var)) {
                elseif ( ( ( ($char = ord($char)) >= 65 && $char <= 90) || ($char >= 97 && $char <= 122) )
                  && $var{1} === ':'
                  && (($char = $var{2}) === '/' || $char === '\\')
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            if ($pathRem) {
              $leMb = drupal_strlen($var = basename($var));
            }
          }
          elseif ($leRaw >= self::$_pathLength) {
            if (DIRECTORY_SEPARATOR == '/') { // *nix; non-Windows.
              if ($var{0} === '/' && drupal_strlen($var = preg_replace(self::$_paths, '', $var)) < $leMb) {
                $pathRem = TRUE;
                $leMb = drupal_strlen($var);
                $var = 'document_root/' . $var;
              }
            }
            else {
              if (drupal_strlen($var = preg_replace(self::$_paths, '', $var)) < $leMb) {
                $pathRem = TRUE;
                $leMb = drupal_strlen($var);
                $var = 'document_root/' . $var;
              }
            }
          }
        }
        //  Escape string, though not if target is file.
        $var = $options['target'] != 'file' ? htmlspecialchars(
          str_replace($options['needles'], $options['replacers'], $var),
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
        ) : str_replace($options['needles'], $options['replacers'], $var);
        //  Truncate string.
        if ($leMb > $trunc) {
          $var = drupal_substr($var, 0, $trunc);
          $truncedTo = $trunc;
        }
        self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $s .= (!$truncedTo ? (!$pathRem ? '' : '|-|!' ) :
            ('|' . $truncedTo . (!$pathRem ? '' : '|!' ) ) )
            . ') '
            . $options['quote_begin'] . (!$pathRem ? '' : '...' ) . $var . (!$truncedTo ? '' : '...') . $options['quote_end']
        );
        return $s;
      case 'resource':
        self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $s = '(resource)' . (!$options['hide_scalars'] ? (' ' . get_resource_type($var) ) : '')
        );
        return $s;
      default: // NULL and unknown type
        self::$_outputLength += strlen( // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $s = '(' . $t . ')'
        );
        return $s;
    }
  }

  /**
   * Logs message from frontend to backend database (or syslog).
   *
   *  Required POST vars:
   *  - (integer) log_no
   *  - (string) type: info | inspect | trace
   *  - (string) category: will always get 'frontend' prefixed (plaintext)
   *  - (string) header (plaintext, max multibyte length 255)
   *  - (string) body (HTML allowed, will be escaped, max multibyte length 100.000)
   *
   * @param string $session_counter
   * @param string $severity
   *  - info | notice | warning | error, etc. must have watchdog severity constant equivalent
   * @return void
   *  - sends 403 header if the expected POST vars arent set or their values arent acceptable
   */
  public static function _logFromFrontend($session_counter, $severity) {
    if (!user_access('inspect frontend log') // Menu should check this, but just in case.
      || !self::_permit('log')
      //  Session counter pattern (cookie inspect__sc + frontend log number): 'ascii-letters-and-numbers:number:number'.
      || !$session_counter || !preg_match('/^[a-zA-Z\d]+\:\d{1,5}\:\d{1,5}\:\d{1,5}$/', $session_counter)
      //  Severity must be a non-empty string that matches a WATCHDOG_ constant.
      || !$severity || strlen($severity) > 32 // Deliberately not drupal_strlen().
        || !defined('WATCHDOG_' . ($severity = strtoupper($severity))) // Deliberately not drupal_strtoupper().
      //  Post-var log_no must be a number, and not larger than 99999.
      || !array_key_exists('log_no', $_POST) || strlen($logNo = '' . $_POST['log_no']) > 5 // Deliberately not drupal_strlen().
        || !is_numeric($logNo)
      //  Post-var type must be a non-empty string, no longer than 32 non-mb chars.
      || !array_key_exists('type', $_POST) || !($le = strlen($type = '' . $_POST['type'])) // Deliberately not drupal_strlen().
        || $le > 32
      //  Post-var category must be a non-empty string, no longer than 32 non-mb chars.
      || !array_key_exists('category', $_POST) || !($le = strlen($category = $_POST['category'])) // Deliberately not drupal_strlen().
        || $le > 32
      //  Post-var header must be a non-empty string, no longer than 255 mb chars.
      || !array_key_exists('header', $_POST)
        || (($lHeader = strlen($header = '' . $_POST['header'])) > 255 && ($lHeader = drupal_strlen($header)) > 255)
      //  Post-var body must be a non-empty string, no longer than 500000 mb chars.
      || !array_key_exists('body', $_POST)
        || (($lBody = strlen($body = trim($_POST['body']))) > 500000 && ($lBody = drupal_strlen($body)) > 500000)
    ) {
      header('HTTP/1.1 403 Forbidden');
      exit;
    }

    if (!self::$_init) {
      self::_init();
    }

    if ($lHeader) {
      $header = htmlspecialchars(
        $header,
        ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
        'UTF-8',
        FALSE // No double encoding
      );
    }
    switch ($type) { // IDE: it _is_ initialized.
      case 'info':
        $header = '[' . $session_counter . '] ' . $header;
        break;
      case 'inspect':
      case 'trace':
        //  Prefix Inspect preface if no header.
        $header = !$lHeader ? ('[frontend Inspect' . ($type == 'inspect' ? '' : ' trace') . ' - ' . $session_counter . ']') :
          ('[' . $session_counter . '] ' . $header);
        break;
      default:
        header('HTTP/1.1 403 Forbidden');
        exit;
    }
    $body = $lBody ? htmlspecialchars(
      $body,
      ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
      'UTF-8',
      FALSE // No double encoding
    ) : '- no message body -';

    $opts = self::$_options['log'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '  class="module-inspect' . ($type == 'inspect' ? '-collapsible inspect-frontend' : '') . '">';
      $tagEnd = '</' . $u . '>';
    }
    //  Truncate message body until not too long.
    $outputMax = self::$_outputMax_log - 1000; // Not Inspect::LOG_MAX_MARGIN, because at this point we only need to consider insert query length.
    $i = 0;
    while (
      (++$i) < 10 // Iteration emergency brake.
      && strlen($message = $tagStart . $header . ($lBody ? $opts['newline'] : '') . $body . $tagEnd) > $outputMax // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
    ) {
      $body = drupal_substr($body, 0, $lBody = floor($lBody / 2));
    }
    try {
      watchdog(
        'frontend ' . htmlspecialchars(
          check_plain($category), // IDE: it _is_ initialized.
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
        ),
        $message, // Escaped previously; is concatenation of escaped header and escaped body.
        NULL,
        constant('WATCHDOG_' . $severity)
      );
    }
    catch (Exception $xc) {
    }
    //  Deliberately dont tell frontend if logging failed.
    $response = new stdClass();
    $response->success = TRUE;
    $response->error = '';
    $response->logNo = $logNo;
    header('Content-Type: application/json; charset=utf-8');
    header('Cache-Control: private, no-store, no-cache, must-revalidate');
    header('Expires: Thu, 01 Jan 1970 00:00:01 GMT');
    //  No reason to use drupal_json_encode(), because we dont send html strings.
    echo json_encode($response);
    flush();
    exit;
  }

  /**
   * Get request time start - in milliseconds - more accurately than the REQUEST_TIME constant.
   *
   * The value is usually slightly higher than REQUEST_TIME, apparantly the latter is a floor'ed value.
   *
   * Uses either $_SERVER['REQUEST_TIME_FLOAT'] (PHP>=5.4), request header X-Request-Received-Processed, or REQUEST_TIME.
   *
   * To enable use of the custom request header (Apache only), add the following to document root .htaccess:
   * @code
   * <IfModule mod_headers.c>
   *   RequestHeader set X-Request-Received-Processing "%t %D"
   * </IfModule>
   * @endcode
   *
   * @return float
   */
  public static function requestTimeMilli() {
    if (!($rtm =& self::$_requestTimeMilli)) {
      $rtm = ($rts = REQUEST_TIME) * 1000;
      if (!empty($GLOBALS['_SERVER']) && array_key_exists('REQUEST_TIME_FLOAT', $_SERVER) && ($v = (int) $_SERVER['REQUEST_TIME_FLOAT'])) {
        $rtm = round($v * 1000, 3);
      }
      elseif (
        ( ( !empty($GLOBALS['_SERVER']) && array_key_exists('HTTP_X_REQUEST_RECEIVED_PROCESSING', $_SERVER)
          && ($v = $_SERVER['HTTP_X_REQUEST_RECEIVED_PROCESSING']) )
        || ( function_exists('apache_request_headers') && ($a = apache_request_headers())
          && array_key_exists('X-Request-Received-Processing', $a) && ($v = $a['X-Request-Received-Processing']) )
        )
        && strlen($v) < 31 && preg_match('/^t\=\d{16,17}\ D\=\d{1,9}/', $v) // Deliberately not drupal_strlen().
      ) {
        $a = explode(' ', $v);
        $proc = (float) substr($a[1], 2); // Deliberately not drupal_substr().
        //  If received time can be reduced to positive integer, and that is close to REQUEST_TIME.
        if ( ($iRec = (int) round(($flRec = (float) substr($a[0], 2)) / 1000000)) > 0 // Deliberately not drupal_strlen().
          && $iRec <= $rts + 1 && $iRec >= $rts - 1
        ) {
          $rtm = round(($flRec + $proc) / 1000, 3);
        }
      }
    }
    return $rtm;
  }

  /**
   * Inspect variable and log to database (or syslog), if permitted.
   *
   *   Like PHP native var_dump()/var_export() except:
   *   - is controlled by permissions
   *   - has much more options, and far better formatting (JSONesque)
   *   - doesnt fail on inspecting $GLOBALS
   *   - aborts if producing too large output, or taking too much time
   *   - doesnt echo, logs to database/syslog
   *
   *   Integer value of $options (depth):
   *   - the value controls how deeply the variable is going to be analyzed
   *   - default: 10 (Inspect::DEPTH_DEFAULT)
   *   - max: 10 (Inspect::DEPTH_MAX)
   *
   *   Array/object value of $options (any number of options):
   *   - (integer) depth (default 10, max 10 ~ Inspect::DEPTH_DEFAULT|DEPTH_MAX)
   *   - (integer) truncate (default 1000, max 10000 ~ Inspect::TRUNCATE_MIN|TRUNCATE_DEFAULT)
   *   - (boolean) hide_scalars: hide values of scalar vars, and type of resource (also supported as value 'hide_scalars')
   *   - (boolean) hide_paths: only relevant for log (paths always hidden for get) (also supported as value 'hide_paths')
   *   - (string) name: '$var_name' or "\$var_name", must be escaped, will be truncated to 255 (default empty, use 'none' to omit)
   *   - (string) message, will be truncated to 255 (default empty)
   *   - (string) category: logging category (default inspect)
   *   - (integer) severity: default WATCHDOG_DEBUG
   *   - (string|array) filter: filter out that|those key name(s) (default empty)
   *   - (array) needles: list of additional char|string needles to be used in str_replace
   *   - (array) replacers: list of additional char|string replacers to be used in str_replace
   *   - (boolean) one_lined: create one-lined formatting (getting only, and also supported as value 'one_lined')
   *   - (boolean) no_folding: dont make output expansible/collapsible (also supported as value 'no_folding')
   *   - (boolean) file_line: display file and line of call to inspect function/method (default do)
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes paths at the beginning of variables:
   *   - always removes document root
   *   - always removes relative and absolute paths when getting (optional when logging)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * Executes variable analysis within try-catch.
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return boolean|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function log($var, $options = NULL) {
    if (!self::_permit('log')) {
      return NULL;
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('log', self::$_options['log'], $options);

    $nl = $opts['newline'];
    try {
      self::$_outputMax = self::$_outputMax_log - static::LOG_MAX_MARGIN;
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch (($errorCode = (int)$xc->getCode())) {
        case Inspect::ERROR_OUTPUTLENGTH: // Output length exceeding maximum.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ((int)$xc1->getCode() != Inspect::ERROR_OUTPUTLENGTH) {
                return '';
              }
              $nspct .= $nl . '(' . $xc1->getCode() . ') ' . $xc1->getMessage();
            }
          }
          break;
        case Inspect::ERROR_EXECTIME: // Max percent of max_execution_time passed.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          break;
        default:
          return FALSE;
      }
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '  class="module-inspect' . (empty($opts['no_folding']) ? '-collapsible' : '') . '">';
      $tagEnd = '</' . $u . '>';
    }
    return self::_log(
        $tagStart
        . ( empty($opts['no_preface']) ?
          ( empty($opts['message']) ?
            ('[Inspect - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . ' - depth:' . $opts['depth'] . ']') :
            ('[' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . '] '
              . $opts['message'] . ':' . $nl . '[Inspect - depth:' . $opts['depth'] . ']')
          ) :
          ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
        )
        . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
        . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), $opts['hide_paths']) . $nl) )
        . $nspct
        . $tagEnd,
        $opts['category'],
        $opts['severity']
    );
  }

  /**
   * Inspect variable and log to file, if permitted.
   *
   *   Array/object value of $options are like Inspect::log()'s options, except:
   *   - (boolean) by_user: log to user specific file instead of common file
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return boolean|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function file($var, $options = NULL) {
    if (!self::_permit('file')) {
      return NULL;
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('file', self::$_options['file'], $options);

    $nl = $opts['newline'];
    try {
      self::$_outputMax = self::$_outputMax_get; // Use getting output max.
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch (($errorCode = (int)$xc->getCode())) {
        case Inspect::ERROR_OUTPUTLENGTH: // Output length exceeding maximum.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ((int)$xc1->getCode() != Inspect::ERROR_OUTPUTLENGTH) {
                return '';
              }
              $nspct .= $nl . '(' . $xc1->getCode() . ') ' . $xc1->getMessage();
            }
          }
          break;
        case Inspect::ERROR_EXECTIME: // Max percent of max_execution_time passed.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          break;
        default:
          return FALSE;
      }
    }
    //  Option no_preface not supported.
    return self::_file(
        ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
          . '[Inspect - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . ' - depth:' . $opts['depth'] . ']'
          . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
          . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), $opts['hide_paths']) . $nl) )
          . $nspct,
        $opts['category'],
        $opts['severity'],
        empty($opts['by_user']) ? FALSE : $opts['by_user']
    );
  }

  /**
   * Inspect variable and get the output as string, if permitted.
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function get($var, $options = NULL) {
    if (!self::_permit('get')) {
      return '';
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('get', self::$_options['get'], $options);

    $nl = $opts['newline'];
    try {
      self::$_outputMax = self::$_outputMax_get;
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch (($errorCode = (int)$xc->getCode())) {
        case Inspect::ERROR_OUTPUTLENGTH: // Output length exceeding maximum.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ((int)$xc1->getCode() != Inspect::ERROR_OUTPUTLENGTH) {
                return '';
              }
              $nspct .= $nl . '(' . $xc1->getCode() . ') ' . $xc1->getMessage();
            }
          }
          break;
        case Inspect::ERROR_EXECTIME: // Max percent of max_execution_time passed.
          $nspct = '(' . $errorCode . ') ' . $xc->getMessage();
          break;
        default:
          return FALSE;
      }
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '  class="module-inspect' . (empty($opts['no_folding']) ? '-collapsible' : '') . '">';
      $tagEnd = '</' . $u . '>';
    }
    if (empty($opts['no_preface'])) {
      $header = empty($opts['message']) ?
        ('[Inspect - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . ' - depth:' . $opts['depth'] . ']') :
        ('[' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . '] '
          . $opts['message'] . ':' . $nl . '[Inspect - depth:' . $opts['depth'] . ']');
    }
    else {
      $header = empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl);
    }
    return $tagStart
      . ( empty($opts['no_preface']) ?
        ( empty($opts['message']) ?
          ('[Inspect - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . ' - depth:' . $opts['depth'] . ']') :
          ('[' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . '] '
            . $opts['message'] . ':' . $nl . '[Inspect - depth:' . $opts['depth'] . ']')
        ) :
        ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
      )
      . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
      . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), TRUE) . $nl) )
      . $nspct . $tagEnd;
  }

  /**
   * Inspect arguments of surrounding function/method and log to database (or syslog), if permitted.
   *
   * $options like Inspect::log().
   *
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return boolean|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function args($options = NULL, $_get = FALSE) {
    $target = !$_get ? 'log' : 'get';
    if (!self::_permit($target)) {
      return $target == 'log' ? NULL : '';
    }
    if (!self::$_init) {
      self::_init();
    }
    $trc = debug_backtrace();
    $frame = NULL;
    for ($i = 0; $i < 4; $i++) {
      if (isset($trc[$i]['class']) && strlen($s = $trc[$i]['class'])) { // Deliberately not drupal_strlen().
        if (strpos($s, 'Inspect') === 0) {
          continue;
        }
      }
      elseif (isset($trc[$i]['function']) && strlen($s = $trc[$i]['function']) // Deliberately not drupal_strlen().
        && ($s == 'inspect' || strpos($s, 'inspect_') === 0)) {
        continue;
      }
      if (isset($trc[$i]['args'])) {
        $frame = $trc[$i];
        break;
      }
    }
    if ($frame) {
      $opts = self::_options($target, self::$_options[$target], $options);
      if (empty($opts['category'])) {
        $opts['category'] = 'inspect args';
      }
      if (empty($opts['name'])) {
        $opts['name'] = (empty($frame['class']) ? '' : ($frame['class'] . (!empty($frame['type']) ? $frame['type'] : '::') ) )
          . (!empty($frame['function']) ? ($frame['function'] . '()') : '' );
      }
      if (!$_get) {
        return self::log($frame['args'], $opts);
      }
      //  If get is one_lined, we want to get rid of the 'array'
      if (($s = self::get($frame['args'], $opts))
        && !empty($opts['one_lined'])) {
        return preg_replace('/(\(\)\s)?\(array\:/', '(', $s, 1);
      }
      return $s;
    }
    return !$_get ? FALSE : '';
  }

  /**
   * Inspect arguments of surrounding function/method and get the output as string, if permitted.
   *
   * Usable for error messages.
   *
   * $options like Inspect::get().
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: 'one_lined' (~ create short single lined format usable for error messages, using depth:1)
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message (except for 'one_lined')
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function args_get($options = 'one_lined') {
    return self::args($options, TRUE);
  }

  /**
   * Inspect constants and log to database (or syslog), if permitted.
   *
   *   $options like Inspect::log(), except:
   *   - (bool) sort: if truthy, sorts constants by name (except if inspecting all defined constants, then only category is sorted)
   *
   * @see Inspect::log()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of that object's class
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - boolean: sort constants
   *   - other type: ignored
   * @return boolean|NULL|string
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   *   - string if truthy arg $_get
   */
  public static function constants($cat_or_instance_or_classname = 'user', $options = NULL, $_get = FALSE) {
    $target = !$_get ? 'log' : 'get';
    if (!self::_permit($target)) {
      return $target == 'log' ? NULL : '';
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options($target, self::$_options[$target], $options, 'depth', 'sort');

    $cat = $cat_or_instance_or_classname;
    $nm = 'constants';
    if (!$cat) {
      $a = get_defined_constants(TRUE);
    }
    elseif (is_string($cat)) {
      if ($cat != 'user') {
        $cat = preg_replace('/[^a-zA-Z\d_\-\ ]/', '', $cat);
      }
      if ($cat != 'user' && substr($cat, 0, 6) === 'class ') { // Deliberately not drupal_substr().
        if (($c = substr($cat, 6)) && class_exists($c)) { // Deliberately not drupal_substr().
          $o = new ReflectionClass($c);
          $a = $o->getConstants();
          $nm = 'class ' . $c . ' ' . $nm;
        }
        else {
          return FALSE;
        }
      }
      else {
        $a = get_defined_constants($cat);
        $a = array_key_exists($cat, $a) ? $a[$cat] : array();
        $nm = $cat . ' ' . $nm;
      }
    }
    elseif (is_object($cat)) {
      $c = get_class($cat);
      $o = new ReflectionClass($c);
      $a = $o->getConstants();
      $nm = 'class ' . $c . ' ' . $nm;
    }
    else { // Bad arg $cat_or_instance_or_classname.
      return !$_get ? FALSE : '';
    }
    if (!empty($opts['sort'])) {
      ksort($a);
    }
    $opts['name'] = $nm;

    return !$_get ? self::log($a, $opts) : self::get($a, $opts);
  }

  /**
   * Inspect constants and get the output as string, if permitted.
   *
   * @see Inspect::get()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of that object's class
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - boolean: sort constants
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function constants_get($cat_or_instance_or_classname = 'user', $options = NULL) {
    return self::constants($cat_or_instance_or_classname, $options, TRUE);
  }

  /**
   * Log a message - and file + line of the call to this method - to database (or syslog), if permitted.
   *
   * Checks message length against maximum allowed database query/syslog length.
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return boolean|NULL
   *  - NULL if user isnt permitted to log inspections
   *  - FALSE: too long message, database query error
   */
  public static function logMessage($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    if (!self::_permit('log')) {
      return NULL;
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts =& self::$_options['log'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '>';
      $tagEnd = '</' . $u . '>';
    }
    return self::_log(
      $tagStart
        . htmlspecialchars(
          str_replace($opts['needles'], $opts['replacers'], $message),
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
        )
        . $opts['newline']
        . self::_fileLine(NULL, FALSE)
        . $tagEnd,
      $category,
      $severity
    );
  }

  /**
   * Log to file, if permitted.
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @param boolean $by_user
   *   - default: FALSE (~ logs to common log file)
   *   - truthy: logs to user specific log file
   * @return boolean|NULL
   *  - NULL if user isnt permitted to log inspections
   *  - FALSE: too long message, or failing to file
   */
  public static function fileMessage($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG, $by_user = FALSE) {
    if (!self::_permit('file')) {
      return NULL;
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts =& self::$_options['file'];
    return self::_file(
      htmlspecialchars(
        str_replace($opts['needles'], $opts['replacers'], $message),
        ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
        'UTF-8',
        FALSE // No double encoding
      ),
      $category,
      $severity,
      $by_user
    );
  }


  //  Utility methods.

  /**
   * Establishes maximum output lengths.
   *
   * If $check or no persistent variables values: sets new values.
   *
   * Maximum for filing is the same as max for getting.
   *
   * Public because module settings dialog calls this method.
   *
   * @param boolean $check
   *  - default: FALSE
   *  - truthy: establishes absolute maxes and returns array of all defaults and maxes
   * @return boolean|array
   */
  public static function outputMax($check = FALSE) {
    $syslog = module_exists('syslog');
    if (!$check
      && (self::$_outputMax_log = variable_get(!$syslog ? 'inspect_outputmax_log' : 'inspect_outputmax_syslog', 0))
      && (self::$_outputMax_get = variable_get('inspect_outputmax_get', 0))
    ) {
      return TRUE;
    }
    //  The rest is fallback, or for admin page module settings form.
    $absolute = 2097152; // 2 Mb
    $logMax = $logDefault = 1048576; // 1 Mb ~ MySql default max_allowed_packet
    $logSyslog = 32768;

    //  Find max string length of db queries (MySql only).
    try {
      if (!($le = (int) db_query('SELECT @@max_allowed_packet')->fetchField())) {
        $le = $logDefault;
      }
      elseif ($le > $absolute) {
        $le = $absolute;
      }
    }
    catch (Exception $xc) { // Possible PDOException.
      self::_log('Failed to read max query length of database', 'inspect', WATCHDOG_WARNING);
      $le = $logDefault;
    }
    $logMax = $le;
    if (!($le = variable_get('inspect_outputmax_log', 0))) {
      variable_set('inspect_outputmax_log', $le = $logMax);
    }
    if (!$syslog) {
      self::$_outputMax_log = $le;
    }

    if (!($le = variable_get('inspect_outputmax_syslog', 0))) {
      variable_set('inspect_outputmax_syslog', $le = $logSyslog);
    }
    if ($syslog) {
      self::$_outputMax_log = $le;
    }

    if (!($le = variable_get('inspect_outputmax_get', 0))) {
      variable_set('inspect_outputmax_get', $le = $absolute);
    }
    self::$_outputMax_get = $le;

    return !$check ? TRUE : array(
      'logDefault' => $logDefault,
      'log' => $logMax,
      'syslog' => $logSyslog,
      'get' => $absolute,
    );
  }

  /**
   * Convert number to string avoiding E-notation for numbers outside system precision range.
   *
   * @param float|integer $num
   * @return string
   */
  public static function numberToString($num) {
    static $precision;
    if (!$precision) {
      $precision = ini_get('precision');
    }
    return ($num < pow(10, $precision) && $num > pow(10, -$precision)) ?  // if within system precision, just string it
      ('' . $num) : number_format($num, 0, '.', '');
  }

  /**
   * Convert number or hex string to any base through 2 and 62  (~ larger than 16).
   *
   * Base 62 produces up to 35% shorter string than base 16.
   *
   *  Like native base_convert() precision fails for numbers (somewhere) larger than (tested on windows box with ini precision 14):
   *  - base 10: 9007199254740991 (> 15 digits)
   *  - base 16: 1fffffffffffff (> 14 digits)
   *
   * @param integer|float|string $num
   * @param integer $fromBase
   *  - default: 10
   *  - only supports 10 and 16 (if other, returns arg $num unaltered)
   * @param integer $toBase
   *  - default (and max): 62
   * @return string
   */
  public static function baseConvert($num, $fromBase = 10, $toBase = 62) {
    $table = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $zero = $table{0};
    $remainder = $num;
    if ($num < 1) {
      if (!$num) {
        return '' . $zero; // zero
      }
      $remainder *= -1;
    }
    if ($toBase == $fromBase) {
      return $remainder;
    }
    switch ($fromBase) {
      case 10:
        break;
      case 16:
        $remainder = base_convert($remainder, 16, 10);
        break;
      default:
        return $remainder;
    }
    $remainder = (float)$remainder; // Arg $num may be string, and base_convert() returns string.
    if ($toBase > 62) {
      $toBase = 62;
    }
    elseif ($toBase < 2) {
      $toBase = 2;
    }
    $res = '';
    //  Find power ~ required number of digits.
    $pow = 0;
    $divisor = 1;
    $limit = 100; // probably way too high, havent seen more than 10 iterations (1fffffffffffff, with precision 14)
    while ($remainder >= ($test = pow($toBase, ++$pow)) && (--$limit)) {
      $divisor = $test;
      --$limit;
    }
    --$pow;
    //  Do modulo procedure.
    $limit = 100; // probably way too high, havent seen more than 14 iterations (1fffffffffffff, with precision 14)
    while ($remainder > 0 && (--$limit)) {
      if ($pow) {
        if ($divisor == $remainder) {
          return $res . $table{1} . str_repeat($zero, $pow);
        }
        elseif ($divisor > $remainder) {
          $res .= $zero;
        }
        else {
          if ($remainder > $divisor) {
            $res .= $table{ (int)floor($remainder / $divisor) };
            $remainder = fmod($remainder, $divisor);
          }
          else {
            $res .= $table{ $pow };
            $remainder = fmod($remainder, $divisor);
          }
          if ($remainder < 1 && !floor($remainder)) {
            return $res . str_repeat($zero, $pow);
          }
        }
      }
      else {
        return $res . $table{ (int)$remainder };
      }
      $divisor = pow($toBase, --$pow);
    }
    return $res;
  }

  /**
   * Secures existance of directory private://module/inspect; optionally subdir to that.
   *
   * @param string $sub_dir
   *  - default: empty
   *  - no starting nor trailing slash
   * @return string|boolean
   *  - FALSE on error
   */
  protected static function _ensureDir($sub_dir = '') {
    if (!($dir = self::$_dir)) {
      if ($dir === NULL) {
        if (!($le = strlen($dir = variable_get('file_private_path', '')))) { // Deliberately not drupal_strlen().
          self::_log(
            'Private files directory not defined, check /admin/config/media/file-system', 'inspect', WATCHDOG_WARNING
          );
          return (self::$_dir = FALSE);
        }
      }
      else {
        return FALSE;
      }
      if ($dir{$le - 1} == '/') { // Remove trailing slash.
        $dir = substr($dir, 0, $le - 1); // Deliberately not drupal_substr().
        $le -= 1;
      }
      if ($le < 2) { // Private files dir cannot be shorter than 2 chars (/x).
        return (self::$_dir = FALSE);
      }
      if ($le < 2
        || (!is_dir($dir .= '/module/inspect') && !mkdir($dir, 0755, TRUE))
      ) {
        self::_log(
          $le < 2 ? 'Private files directory cannot be shorter than 2 chars, check /admin/config/media/file-system' :
            'Directory private://module/inspect is not a dir or cannot be created',
          'inspect',
          WATCHDOG_WARNING
        );
        return (self::$_dir = FALSE);
      }
      self::$_dir = $dir;
    }
    if (!$sub_dir
      || is_dir($dir .= '/' . $sub_dir) || mkdir($dir, 0755, TRUE)
    ) {
      return $dir;
    }
    self::_log(
      'Directory private://module/inspect/' . $sub_dir . ' is not a dir or cannot be created',
      'inspect',
      WATCHDOG_WARNING
    );
    return FALSE; // Failed to ensure subdir.
  }

  /**
   * Get file and line of outmost call to an inspect function or public method.
   *
   * @param array|NULL $trace
   *  - slightly cheaper when given debug_backtrace() arg
   * @param boolean $hidePaths
   *  - default: TRUE
   * @return string
   */
  protected static function _fileLine($trace = NULL, $hidePaths = TRUE) {
    if ($trace) {
      if (($le = count($trace)) > 3) {
        $le = 3;
      }
    }
    elseif (($le = count($trace = debug_backtrace())) > 4) {
      $le = 4;
    }
    //  finds last frame having an Inspect class or inspect function
    $iFrm = -1;
    for ($i = 0; $i < $le; $i++) {
      $nspct = FALSE;
      if (isset($trace[$i]['class']) && strlen($s = $trace[$i]['class'])) { // Deliberately not drupal_strlen().
        $nspct = (strpos(strtolower($s), 'inspect') === 0);
      }
      else {
        $nspct = (isset($trace[$i]['function']) && strlen($s = $trace[$i]['function']) // Deliberately not drupal_strlen().
          && ($s == 'inspect' || strpos($s, 'inspect_') === 0) );
      }
      if ($nspct) {
        $iFrm = $i;
      }
      elseif ($iFrm > -1) { // Break a first non-inspect method/function after an inspect method/function was found.
        break;
      }
    }
    return '@'
      . ($iFrm > -1 ?
        (
          (isset($trace[$iFrm]['file']) && strlen($s = trim($trace[$iFrm]['file'])) ? // Deliberately not drupal_strlen().
            ($hidePaths ? basename($s) : ('document_root/' . preg_replace(self::$_paths, '', $s)) ) :
            'unknown'
          )
          . ':'
          . (isset($trace[$iFrm]['line']) ? trim($trace[$iFrm]['line']) : '?')
        ) : 'unknown');
  }

  /**
   * Checks message length against maximum allowed database query/syslog length.
   *
   * Calling method must guarantee to escape message (whereas category will be escaped here).
   *
   * @see Inspect::logMessage()
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return boolean
   *  - FALSE: too long message, database query error
   */
  protected static function _log($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    try {
      $outputMax = self::$_outputMax_log - 1000; // Not Inspect::LOG_MAX_MARGIN, because at this point we only need to consider insert query length.
      if (strlen($message) > $outputMax) { // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
        $message = 'Message length [' . strlen($message) // Deliberately not drupal_strlen().
          . '] exceeds max allowed [' . $outputMax . ']';
        $severity = $severity > -1 && $severity < WATCHDOG_WARNING ? $severity : WATCHDOG_WARNING;
        $return = FALSE;
      }
      else {
        $return = TRUE;
      }
      watchdog(
        check_plain(drupal_substr($category, 0, 64)), // Will generate error when extracting strings for translation, but cant do anything about.
        $message, // Will generate error when extracting strings for translation, but doing replacing just the message doesnt make sense.
        NULL,
        $severity > -1 ? $severity : WATCHDOG_DEBUG
      );
      return $return;
    }
    catch (Exception $xc) {
    }
    return FALSE;
  }

  /**
   * Checks message length against maximum allowed file length.
   *
   * Calling method must guarantee to escape message (whereas category will be escaped here).
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @param boolean $by_user
   *   - default: FALSE (~ logs to common log file)
   *   - truthy: logs to user specific log file
   * @return boolean
   *  - FALSE: too long message, or failing to file
   */
  protected static function _file($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG, $by_user = FALSE) {
    if (!($dir = self::$_logDir) && !(self::$_logDir = $dir = self::_ensureDir('logs'))) {
      return FALSE;
    }
    $category = check_plain(drupal_substr($category, 0, 64));

    $outputMax = self::$_outputMax_get; // Use get max for file.
    if (strlen($message) > $outputMax) { // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
      try {
        watchdog(
          $category,
          'Message length [' . strlen($message) // Deliberately not drupal_strlen().
            . '] exceeds max allowed [' . $outputMax . ']',
          NULL,
          $severity > -1 && $severity < WATCHDOG_WARNING ? $severity : WATCHDOG_WARNING
        );
      }
      catch (Exception $xc) {
      }
      return FALSE;
    }
    //  Stringify severity.
    switch ((int) $severity) {
      case WATCHDOG_DEBUG:
        $sev = 'debug';
        break;
      case WATCHDOG_ALERT:
        $sev = 'alert';
        break;
      case WATCHDOG_CRITICAL:
        $sev = 'critical';
        break;
      case WATCHDOG_ERROR:
        $sev = 'error';
        break;
      case WATCHDOG_WARNING:
        $sev = 'warning';
        break;
      case WATCHDOG_NOTICE:
        $sev = 'notice';
        break;
      case WATCHDOG_INFO:
      case -1:
        $sev = 'info';
        break;
      default: // 0
        $sev = 'emergency';
    }
    $uid = $GLOBALS['user']->uid;
    $rtm = self::$_requestTimeMilli;
    return file_put_contents(
      $dir . '/inspect_' . date('Ymd', (int) round($rtm / 1000))
        . (!$by_user ? '' : // For anonymous user: avoid filing concurrency by using the session number.
          ('_user_' . $uid . ($uid ? '' : ('_' . self::$_sessionCounters[0])))
        )
        . '.log',
      join(':', self::$_sessionCounters)
        . ' - ' . date('Y-m-d H:i:s', floor($rtm / 1000)) . ' - user: ' . $uid
        . ' --------------------------------------------------'
        . "\n" . $_SERVER['REQUEST_METHOD'] . ' ' . request_uri()
        . "\nseverity+category: " . $sev . ' - ' . $category
        . "\n" . $message . "\n\n",
      FILE_APPEND
    ) ? TRUE : FALSE;
  }


  //  Deprecated.

  /**
   * Obsolete since 7.x-4.0, logs error - and the method will be removed some day.
   *
   * @deprecated
   * @see Inspect::logMessage()
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return boolean
   *  - FALSE
   */
  public static function log_db($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    //  @todo: remove this method from some version > 7.x-4.0, obsolete since that version. Remove when few use <4.0.
    watchdog(
      'inspect',
      'Use Inspect::logMessage() instead - method Inspect::log_db() is obsolete since Inspect v. 7.x-4.0, and will be removed some day',
      NULL,
      WATCHDOG_ERROR
    );
    return FALSE;
  }

  /**
   * Obsolete since 7.x-4.0, logs error - and the method will be removed some day.
   *
   * @deprecated
   * @see Inspect::fileMessage()
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @param boolean $by_user
   *   - default: FALSE (~ logs to common log file)
   *   - truthy: logs to user specific log file
   * @return boolean
   *  - FALSE
   */
  public static function log_file($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG, $by_user = FALSE) {
    //  @todo: remove this method from some version > 7.x-4.0, obsolete since that version. Remove when few use <4.0.
    watchdog(
      'inspect',
      'Use Inspect::fileMessage() instead - method Inspect::log_file() is obsolete since Inspect v. 7.x-4.0, and will be removed some day',
      NULL,
      WATCHDOG_ERROR
    );
    return FALSE;
  }
}